#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat.zmi.viewlet.notifications module

This module defines a viewlet which can be used to display user notifications.
"""

from pyams_chat.zmi.viewlet.interfaces import IUserNotificationsViewletManager
from pyams_security.interfaces.names import UNKNOWN_PRINCIPAL_ID
from pyams_template.template import template_config
from pyams_viewlet.manager import TemplateBasedViewletManager, WeightOrderedViewletManager, \
    viewletmanager_config
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.viewlet import IUserLinksViewletManager


__docformat__ = 'restructuredtext'


@viewletmanager_config(name='pyams.usernotifications', layer=IAdminLayer,
                       manager=IUserLinksViewletManager, weight=850,
                       provides=IUserNotificationsViewletManager)
@template_config(template='templates/notifications.pt')
class UserNotificationsViewlet(TemplateBasedViewletManager, WeightOrderedViewletManager):
    """User notifications viewlet manager"""

    def __new__(cls, context, request, view, manager):  # pylint: disable=unused-argument
        principal = request.principal
        if principal.id == UNKNOWN_PRINCIPAL_ID:
            return None
        return WeightOrderedViewletManager.__new__(cls)

    render_empty = True

    @property
    def jwt_refresh_route(self):
        """JWT authentication endpoint"""
        return self.request.registry.settings.get('pyams_chat.jwt_refresh_route',
                                                  '/api/auth/jwt/token')

    @property
    def jwt_verify_route(self):
        """JWT token verification route"""
        return self.request.registry.settings.get('pyams_chat.jwt_verify_route',
                                                  '/api/auth/jwt/verify')

    @property
    def ws_endpoint(self):
        """Websocket endpoint"""
        return self.request.registry.settings.get('pyams_chat.ws_endpoint')



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS chat.include module

This module is used for Pyramid integration
"""

import re

from pyramid.settings import asbool

from pyams_chat.client import RedisClient
from pyams_chat.interfaces import CHAT_PING_ROUTE, CHAT_WORKER_ROUTE, IRedisClient, \
    REST_CONTEXT_ROUTE, \
    REST_NOTIFICATIONS_ROUTE


__docformat__ = 'restructuredtext'


def client_from_config(settings, prefix='pyams_chat.'):
    """
    Instantiate and configure a Redis client from settings.

    In a typical Pyramid application, just include ``pyams_chat`` and use
    :py:func:`get_client` function to get access to the shared :py:class:`.client.RedisClient`
    instance (which is also available using *request.redis_client* notation).
    """
    return RedisClient(
        server_url=settings.get(f'{prefix}redis_server', 'redis://localhost:6379/0'),
        use_transaction=settings.get(f'{prefix}use_transaction', True)
    )


def get_client(request):
    """
    Get registered Redis client
    """
    registry = request.registry
    return registry.queryUtility(IRedisClient)


def include_package(config):
    """Pyramid package include"""

    registry = config.registry
    settings = registry.settings

    # add translations
    config.add_translation_dirs('pyams_chat:locales')

    # add request methods
    config.add_request_method(get_client, 'redis_client', reify=True)

    # register ServiceWorker script route
    config.add_route(CHAT_PING_ROUTE,
                     settings.get('pyams_chat.chat_ping_route',
                                  '/chat-ping'))

    config.add_route(CHAT_WORKER_ROUTE,
                     settings.get('pyams_chat.chat_worker_route',
                                  '/chat-sw.js'))

    # register new REST API routes
    config.add_route(REST_CONTEXT_ROUTE,
                     settings.get('pyams_chat.rest_context_route',
                                  '/api/chat/context'))

    config.add_route(REST_NOTIFICATIONS_ROUTE,
                     settings.get('pyams_chat.rest_notifications_route',
                                  '/api/chat/notifications'))

    # initialize Redis client
    if asbool(settings.get('pyams_chat.start_client', True)):
        client = client_from_config(settings)
        registry.registerUtility(client, IRedisClient)

    # package scan
    try:
        import pyams_zmi  # pylint: disable=import-outside-toplevel,unused-import
    except ImportError:
        config.scan(ignore=[re.compile(r'pyams_chat\..*\.zmi\.?.*').search])
    else:
        config.scan()



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat.zmi.worker module

This module is used to declare specific endpoints which are required to
use chat service worker.
"""

import os

from pyramid.response import FileResponse, Response
from pyramid.view import view_config

from pyams_chat.interfaces import CHAT_PING_ROUTE, CHAT_WORKER_ROUTE


__docformat__ = 'restructuredtext'


@view_config(route_name=CHAT_PING_ROUTE)
def chat_ping(request):
    """Chat ping endpoint"""
    return Response(body='PONG')


@view_config(route_name=CHAT_WORKER_ROUTE)
def chat_worker_script(request):
    """Chat ServiceWorker route"""
    here = os.path.dirname(__file__)
    script = os.path.join(here, 'resources', 'js', 'chat-sw.js')
    return FileResponse(script, request=request)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat.api module

This module provides REST API for chat services.
"""

import json
from datetime import datetime
import sys

from colander import MappingSchema, OneOf, SchemaNode, String, drop
from cornice import Service
from pyramid.httpexceptions import HTTPAccepted, HTTPBadRequest, HTTPForbidden, HTTPNotFound, \
    HTTPOk, HTTPServiceUnavailable, HTTPUnauthorized

from pyams_chat.interfaces import IChatMessageHandler, REST_CONTEXT_ROUTE, \
    REST_NOTIFICATIONS_ROUTE
from pyams_chat.message import ChatMessage


__docformat__ = 'restructuredtext'

from pyams_chat import _


TEST_MODE = sys.argv[-1].endswith('/test')


class StatusSchema(MappingSchema):
    """Base status response schema"""
    status = SchemaNode(String(),
                        title=_("Response status"),
                        validator=OneOf(('success', 'error')))


class ContextSchema(StatusSchema):
    """User context schema"""
    principal = SchemaNode(String(),
                           title=_("Principal ID"))


class ErrorSchema(MappingSchema):
    """Base error schema"""
    status = SchemaNode(String(),
                        title=_("Response status"))
    message = SchemaNode(String(),
                         title=_("Error message"),
                         missing=drop)


jwt_responses = {
    HTTPOk.code: ContextSchema(description=_("User context properties")),
    HTTPAccepted.code: StatusSchema(description=_("Token accepted")),
    HTTPNotFound.code: ErrorSchema(description=_("Page not found")),
    HTTPUnauthorized.code: ErrorSchema(description=_("Unauthorized")),
    HTTPForbidden.code: ErrorSchema(description=_("Forbidden access")),
    HTTPBadRequest.code: ErrorSchema(description=_("Missing arguments")),
    HTTPServiceUnavailable.code: ErrorSchema(description=_("Service unavailable"))
}

if TEST_MODE:
    service_params = {}
else:
    service_params = {
        'response_schemas': jwt_responses
    }


chat_service = Service(name=REST_CONTEXT_ROUTE,
                       pyramid_route=REST_CONTEXT_ROUTE,
                       description="PyAMS chat context API")


@chat_service.get(require_csrf=False,
                  **service_params)
def get_chat_context(request):
    """REST chat context service"""
    principal = request.principal
    message = ChatMessage.create_empty_message(request)
    adapters = [
        name
        for name, adapter in request.registry.getAdapters((message, ),
                                                          IChatMessageHandler)
    ]
    return {
        'status': 'success',
        'principal': {
            'id': principal.id,
            'title': principal.title,
            'principals': tuple(request.effective_principals)
        },
        'context': {
            '*': adapters
        }
    }


notifications_service = Service(name=REST_NOTIFICATIONS_ROUTE,
                                pyramid_route=REST_NOTIFICATIONS_ROUTE,
                                description='PyAMS chat notifications API')


@notifications_service.get(require_csrf=False,
                           **service_params)
def get_notifications(request):
    """REST notifications service"""

    def filter_messages(messages):
        """Filter user notifications"""
        for message in messages:
            if isinstance(message, (str, bytes)):
                message = json.loads(message)
            # don't get messages from other hosts
            if message.get('host') != request.host_url:
                continue
            # don't get messages from current user
            if message.get('source', {}).get('id') == request.principal.id:
                continue
            # filter message targets
            target = message.get('target', {})
            if set(request.effective_principals) & set(target.get('principals', ())):
                yield message

    timestamp = datetime.utcnow().timestamp()
    client = request.redis_client
    if client is None:
        return {
            "timestamp": timestamp,
            "notifications": []
        }
    settings = request.registry.settings
    notifications_key = settings.get('pyams_chat.notifications_key', 'chat:notifications')
    notifications = client.redis.lrange(f'{notifications_key}::{request.host_url}', 0, -1)
    return {
        "timestamp": timestamp,
        "notifications": list(filter_messages(notifications or ()))
    }



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat.client module

This module defines a Redis transaction manager
"""

import json
import logging
import sys

import transaction

try:
    from fakeredis import FakeRedis
except ImportError:
    FakeRedis = None

from redis import Redis

from pyams_utils.transaction import TransactionClient, transactional


__docformat__ = 'restructuredtext'


LOGGER = logging.getLogger('PyAMS (chat)')

TEST_MODE = sys.argv[-1].endswith('/test')


class RedisClient(TransactionClient):
    """Redis client class"""

    def __init__(self,
                 server_url=None,
                 use_transaction=True,
                 transaction_manager=transaction.manager):
        super().__init__(use_transaction, transaction_manager)
        assert server_url is not None, "You must provide a Redis connection URL!"
        factory = FakeRedis if TEST_MODE else Redis
        self.redis = factory.from_url(server_url)

    def close(self):
        """Close redis connection"""
        self.redis.close()

    def __enter__(self):
        return self.redis

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.redis.close()

    @transactional
    def publish(self, channel, message):
        """Public data to given channel"""
        if not isinstance(message, str):
            message = json.dumps(message)
        LOGGER.debug(f'Publishing message to {channel} channel...')  # pylint: disable=logging-fstring-interpolation
        self.redis.publish(channel, message)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat.handler.login module

This module defines a custom *login* handler, which is used to send notifications
on any user login to site managers.
"""

from pyramid.events import subscriber

from pyams_chat.interfaces import IChatMessage, IChatMessageHandler
from pyams_chat.message import ChatMessage
from pyams_security.interfaces import IProtectedObject
from pyams_security.interfaces.names import ADMIN_USER_ID, SYSTEM_ADMIN_ROLE
from pyams_security.interfaces.plugin import IAuthenticatedPrincipalEvent
from pyams_security.utility import get_principal
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.request import query_request


__docformat__ = 'restructuredtext'

from pyams_chat import _  # pylint: disable=ungrouped-imports


@subscriber(IAuthenticatedPrincipalEvent)
def handle_authenticated_principal(event):
    """Handle authenticated principal"""
    request = query_request()
    if request is None:
        return
    principal = get_principal(request, principal_id=event.principal_id)
    translate = request.localizer.translate
    message = ChatMessage(request=request,
                          action='notify',
                          category='user.login',
                          source=principal.id,
                          title=translate(_("User login")),
                          message=translate(_("{} logged in...")).format(principal.title))
    message.send()


@adapter_config(name='user.login',
                required=IChatMessage,
                provides=IChatMessageHandler)
class UserLoginMessageTarget(ContextAdapter):
    """User login message target adapter"""

    def get_target(self):
        """Get message targets"""
        principals = {ADMIN_USER_ID}
        root = self.context.request.root
        protection = IProtectedObject(root, None)
        if protection is not None:
            principals |= protection.get_principals(SYSTEM_ADMIN_ROLE)
        return {
            'principals': tuple(principals)
        }



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat.zmi module

This module defines package resources and viewlets.
"""

__docformat__ = 'restructuredtext'

from fanstatic import Library, Resource


library = Library('pyams_chat', 'resources')

chat_js = Resource(library, 'js/pyams_chat.js',
                   minified='js/pyams_chat.min.js')



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat.interfaces module

This module defines main chat interfaces.
"""

from zope.interface import Attribute, Interface
from zope.schema import Dict, Text, TextLine, URI

from pyams_chat import _


CHAT_PING_ROUTE = 'chat.ping'
CHAT_WORKER_ROUTE = 'chat.worker'

REST_CONTEXT_ROUTE = 'chat.rest.context'
REST_NOTIFICATIONS_ROUTE = 'chat.rest.notifications'


class IChatMessage(Interface):
    """Chat message interface"""

    host = TextLine(title=_("Message host name"))

    channel = TextLine(title=_("Message channel"))

    action = TextLine(title=_("Message action"))

    category = TextLine(title=_("Message category"))

    status = TextLine(title=_("Message status"))

    title = TextLine(title=_("Message title"))

    message = Text(title=_("Message content"))

    image = URI(title=_("Message image URL"),
                required=False)

    source = Dict(title=_("Message source"),
                  description=_("Attributes of the principal which emitted the notification"),
                  key_type=TextLine(),
                  value_type=TextLine())

    target = Dict(title=_("Message target"),
                  description=_("Message targets can be principals, roles..."),
                  key_type=TextLine())

    url = TextLine(title=_("Notification URL"),
                   description=_("URL target of this message notification"),
                   required=False)

    timestamp = TextLine(title=_("Message timestamp in ISO-8601 format"))

    user_data = Dict(title=_("Internal user data"),
                     description=_("User data used for internal usage; these data are not "
                                   "sent in notifications contents"))

    def send(self):
        """Send notification to recipients"""


class IChatMessageExtension(Interface):
    """Chat message extension interface

    These extensions, defined as adapters, can be used to add custom
    information to an existing message.

    This interface doesn't provide any method, as adapters should
    directly extend the original message.
    """

    weight = Attribute("Optional ordering weight")


class IChatMessageHandler(Interface):
    """Chat message handler interface

    This interface is used to define adapters which can be used to
    define a message target based on it's category.
    """

    weight = Attribute("Optional ordering weight")

    def get_target(self):
        """Get message target"""


class IRedisClient(Interface):
    """Redis client interface for chat"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_chat.message module

This message defines main chat message class.
"""

import json
from datetime import datetime, timezone
from json import JSONEncoder

from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty

from pyams_chat.include import get_client
from pyams_chat.interfaces import IChatMessage, IChatMessageExtension, IChatMessageHandler
from pyams_file.image import get_image_selection
from pyams_security.interfaces.base import IPrincipalInfo
from pyams_security.utility import get_principal
from pyams_utils.adapter import get_adapter_weight

try:
    from pyams_zmi.interfaces.configuration import IZMIConfiguration
    from pyams_zmi.interfaces.profile import IUserProfile
except ImportError:
    IZMIConfiguration = None
    IUserProfile = None


__docformat__ = 'restructuredtext'


@implementer(IChatMessage)
class ChatMessage:  # pylint: disable=too-many-instance-attributes
    """Chat message object"""

    host = FieldProperty(IChatMessage['host'])
    channel = FieldProperty(IChatMessage['channel'])
    action = FieldProperty(IChatMessage['action'])
    category = FieldProperty(IChatMessage['category'])
    status = FieldProperty(IChatMessage['status'])
    title = FieldProperty(IChatMessage['title'])
    message = FieldProperty(IChatMessage['message'])
    image = FieldProperty(IChatMessage['image'])
    source = FieldProperty(IChatMessage['source'])
    target = FieldProperty(IChatMessage['target'])
    url = FieldProperty(IChatMessage['url'])
    timestamp = FieldProperty(IChatMessage['timestamp'])
    user_data = FieldProperty(IChatMessage['user_data'])

    @classmethod
    def create_empty_message(cls, request):
        """Create new empty message"""
        return ChatMessage(request,
                           action='',
                           category='',
                           title='',
                           message='')

    def __init__(self, request, **settings):
        self.request = request
        self.host = settings.pop('host', request.host_url)
        self.channel = settings.pop('channel',
                                    request.registry.settings.get('pyams_chat.channel_name'))
        self.context = settings.pop('context', request.context)
        self.action = settings.pop('action', None)
        self.category = settings.pop('category', None)
        self.status = settings.pop('status', 'info')
        self.title = settings.pop('title', None)
        self.message = settings.pop('message', None)
        self.image = settings.pop('image', None)
        source = settings.pop('source', None) or request.principal.id
        if IPrincipalInfo.providedBy(source):  # pylint: disable=no-value-for-parameter
            source = source.id
        self.source_id = source
        self.url = settings.pop('url', None)
        self.timestamp = datetime.now(timezone.utc).isoformat()
        self.user_data = settings

    def get_source(self, principal_id=None):
        """Get message source"""
        if principal_id is None:
            principal_id = self.request.principal.id
        principal = get_principal(self.request, principal_id)
        self.source = {
            'id': principal.id,
            'title': principal.title
        }
        for _adapter in sorted(
                self.request.registry.getAdapters((self,), IChatMessageExtension),
                key=get_adapter_weight):
            pass  # adapters should automatically extend current message
        # check user profile
        avatar = None
        if IUserProfile is not None:
            profile = IUserProfile(principal, None)
            if profile is not None:
                avatar = profile.get_avatar(size='128x128')
        if (avatar is None) and (IZMIConfiguration is not None):
            configuration = IZMIConfiguration(self.request.root, None)
            if configuration is not None:
                avatar = get_image_selection(configuration.favicon, 'square', '128x128',
                                             self.request)
        if avatar is not None:
            self.source['avatar'] = avatar

    def get_target(self):
        """Get message target"""
        target = {}
        handler = self.request.registry.queryAdapter(self, IChatMessageHandler,
                                                     name=self.category)
        if handler is not None:
            target = handler.get_target()
        self.target = target

    def send(self):
        """Get message source and target and publish to Redis"""
        self.get_source(self.source_id)
        self.get_target()
        self.publish()

    def publish(self):
        """Convert message to JSON and publish to Redis channel"""
        client = getattr(self.request, 'redis_client', None)
        if client is None:
            client = get_client(self.request)
        if client is not None:
            json_data = json.dumps(self, cls=ChatMessageEncoder)
            client.publish(self.channel, json_data)


class ChatMessageEncoder(JSONEncoder):
    """Chat message encoder"""

    def default(self, obj):  # pylint: disable=arguments-renamed
        if isinstance(obj, ChatMessage):
            return {
                'host': obj.host,
                'channel': obj.channel,
                'action': obj.action,
                'category': obj.category,
                'status': obj.status,
                'title': obj.title,
                'message': obj.message,
                'image': obj.image,
                'source': obj.source,
                'target': obj.target,
                'url': obj.url,
                'timestamp': obj.timestamp
            }
        return super().default(obj)
